/**
 * The internal recursive functions used by the AVLTree class, and additional
 * helper functions. These perform outside of the tree class context, and take
 * nodes to operate on. This makes them reusable for many more applications.
 *
 * While the AVLTree interface only exposes its contained values the functions
 * here return nodes with these values within, so that internally we can easily
 * traverse subtrees and otherwise operate on them.
 *
 * @module Internal - Functions
 */
import type { Ord } from '../avl-bst';
import { AVLNode as Node } from './AVLNode';
export declare const enum Ordering {
    LT = -1,
    EQ = 0,
    GT = 1
}
/**
 * Returns the height of given node or zero if the node is `null`.
 *
 * @category Helper
 */
export declare function nodeHeight(node: Node | null): number;
/**
 * Updates the given node's height property based on the current heights
 * of its left and right subtrees.
 *
 * @category Helper
 */
export declare function updateNodeHeight(node: Node): void;
/**
 * Updates the given node's size property based on the current sizes
 * of its left and right subtrees.
 *
 * @category Helper
 */
export declare function updateNodeSize(node: Node): void;
/**
 * Returns the given node's balance: the relationship between the heights
 * of its left and right subtrees. Returns a negative number if the node's
 * left child is the higher height child, or a positive number if the right
 * child is the higher height child. Returns zero if either the node is `null`
 * or both its children have equal height.
 *
 * @category Helper
 */
export declare function nodeBalance(node: Node | null): number;
/**
 * Takes two values of type `K extends Orderable` and returns their ordering.
 *
 * @category Helper
 */
export declare function scalarCompare<K extends Ord>(ka: K, kb: K): Ordering;
/**
 * Writes the key and value (by reference) from the given `from` node to the given `to` node.
 *
 * @category Helper
 */
export declare function writeNodeContents<K extends Ord, V>({ from, to }: {
    from: Node<K, V>;
    to: Node<K, V>;
}): void;
/**
 * Returns the node within the subtree of given node that ranks lowest.
 * Returns the given node itself if it has no children, or `null`
 * if the given node is itself `null`.
 *
 * @category Tree recursion
 */
export declare function minNode<K extends Ord, V>(node: Node<K, V> | null): Node<K, V> | null;
/**
 * Returns the node within the subtree of given node that ranks highest.
 * Returns the given node itself if it has no children, or `null`
 * if the given node is itself `null`.
 *
 * @category Tree recursion
 */
export declare function maxNode<K extends Ord, V>(node: Node<K, V> | null): Node<K, V> | null;
/**
 * Performs in-order traversal of the given node's subtree, applying the given
 * `fn` to each contained value.
 * @category Tree recursion
 */
export declare function traverseInOrder<K extends Ord, V>(node: Node<K, V> | null, fn: (node: Node<K, V>) => void): void;
/**
 * Performs pre-order traversal of the given node's subtree, applying the given
 * `fn` to each contained value.
 * @category Tree recursion
 */
export declare function traversePreOrder<K extends Ord, V>(node: Node<K, V> | null, fn: (node: Node<K, V>) => void): void;
/**
 * Performs post-order traversal of the given node's subtree, applying the given
 * `fn` to each contained value.
 * @category Tree recursion
 */
export declare function traversePostOrder<K extends Ord, V>(node: Node<K, V> | null, fn: (node: Node<K, V>) => void): void;
/**
 * Folds (reduces) the given node's subtree left-to-right using in-order traversal.
 *
 * @category Tree recursion
 */
export declare function foldLeft<K extends Ord, V, T>(node: Node<K, V> | null, fn: (acc: T, node: Node<K, V>) => T, seed: T): T;
/**
 * Folds (reduces) the given node's subtree right-to-left using reversed in-order traversal.
 *
 * @category Tree recursion
 */
export declare function foldRight<K extends Ord, V, T>(node: Node<K, V> | null, fn: (acc: T, node: Node<K, V>) => T, seed: T): T;
/**
 * Search a node by a given `searchKey`, matching against keys of nodes. Returns the
 * matching node if found, or `null` otherwise.
 *
 * @category Tree recursion
 */
export declare function search<K extends Ord, V>(node: Node<K, V> | null, searchKey: K): Node<K, V> | null;
/**
 * Inserts the given key an corresponding value into the sub-tree of the given node.
 * Returns an array of length 2 with at:
 *
 * - `[0]` - the node that takes the place of given node after insertion and re-balancing,
 *   which might also be the given node itself.
 *
 * - `[1]` - A boolean indicating whether the new key and value were actually inserted.
 *   This will be `false` if the key was already in the tree prior to insertion.
 *
 * @category Tree recursion
 */
export declare function insert<K extends Ord, V>(key: K, val: V, node: Node<K, V> | null): [Node<K, V>, boolean];
/**
 * Deletes the node by given key from the subtree of given node. Returns an array with:
 *
 * - `[0]` - the node that takes the place of given node after deletion and re-balancing,
 *   which might also be the given node itself.
 *
 * - `[1]` - A boolean indicating whether a node was actually deleted.
 *   This will be `false` if the key was not found in the tree.
 *
 * @category Tree recursion
 */
export declare function deleteKey<K extends Ord, V>(key: K, node: Node<K, V> | null): [Node<K, V> | null, boolean];
/**
 * Performs a left rotation on the given node. Returns the node that takes the
 * place of the given node after rotation.
 *
 * ```text
 * ┌──────────────────────────────────────────────────────────────────┐
 * │ argument -> z                                      y <- return   │
 * │            /  \                                  /   \           │
 * │           T1   y         rotateLeft(z)          z      x         │
 * │               /  \       - - - - - - - ->      / \    / \        │
 * │              T2   x                           T1  T2 T3  T4      │
 * │                  / \                                             │
 * │                T3  T4                                            │
 * └──────────────────────────────────────────────────────────────────┘
 * ```
 * @category Helper
 */
export declare function rotateLeft<K extends Ord, V>(node: Node<K, V>): Node<K, V>;
/**
 * Performs a right rotation on the given node. Returns the node that takes the
 * place of the given node after rotation.
 *
 * ```text
 * ┌──────────────────────────────────────────────────────────────────┐
 * │ argument -> z                                      y <- return   │
 * │            / \                                   /   \           │
 * │           y   T4      rotateRight(z)            x      z         │
 * │          / \          - - - - - - - - ->      /  \    /  \       │
 * │         x   T3                               T1  T2  T3  T4      │
 * │        / \                                                       │
 * │      T1   T2                                                     │
 * └──────────────────────────────────────────────────────────────────┘
 * ```
 * @category Helper
 */
export declare function rotateRight<K extends Ord, V>(node: Node<K, V>): Node<K, V>;
