"use strict";
/**
 * The internal recursive functions used by the AVLTree class, and additional
 * helper functions. These perform outside of the tree class context, and take
 * nodes to operate on. This makes them reusable for many more applications.
 *
 * While the AVLTree interface only exposes its contained values the functions
 * here return nodes with these values within, so that internally we can easily
 * traverse subtrees and otherwise operate on them.
 *
 * @module Internal - Functions
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.rotateRight = exports.rotateLeft = exports.deleteKey = exports.insert = exports.search = exports.foldRight = exports.foldLeft = exports.traversePostOrder = exports.traversePreOrder = exports.traverseInOrder = exports.maxNode = exports.minNode = exports.writeNodeContents = exports.scalarCompare = exports.nodeBalance = exports.updateNodeSize = exports.updateNodeHeight = exports.nodeHeight = void 0;
var AVLNode_1 = require("./AVLNode");
/**
 * Returns the height of given node or zero if the node is `null`.
 *
 * @category Helper
 */
function nodeHeight(node) {
    var _a;
    return (_a = node === null || node === void 0 ? void 0 : node.height) !== null && _a !== void 0 ? _a : 0;
}
exports.nodeHeight = nodeHeight;
/**
 * Updates the given node's height property based on the current heights
 * of its left and right subtrees.
 *
 * @category Helper
 */
function updateNodeHeight(node) {
    node.height = 1 + Math.max(nodeHeight(node.left), nodeHeight(node.right));
}
exports.updateNodeHeight = updateNodeHeight;
/**
 * Updates the given node's size property based on the current sizes
 * of its left and right subtrees.
 *
 * @category Helper
 */
function updateNodeSize(node) {
    var _a, _b, _c, _d;
    node.size = 1 + ((_b = (_a = node.left) === null || _a === void 0 ? void 0 : _a.size) !== null && _b !== void 0 ? _b : 0) + ((_d = (_c = node.right) === null || _c === void 0 ? void 0 : _c.size) !== null && _d !== void 0 ? _d : 0);
}
exports.updateNodeSize = updateNodeSize;
/**
 * Returns the given node's balance: the relationship between the heights
 * of its left and right subtrees. Returns a negative number if the node's
 * left child is the higher height child, or a positive number if the right
 * child is the higher height child. Returns zero if either the node is `null`
 * or both its children have equal height.
 *
 * @category Helper
 */
function nodeBalance(node) {
    return node ? nodeHeight(node.right) - nodeHeight(node.left) : 0;
}
exports.nodeBalance = nodeBalance;
/**
 * Takes two values of type `K extends Orderable` and returns their ordering.
 *
 * @category Helper
 */
function scalarCompare(ka, kb) {
    return ka < kb ? -1 /* LT */ :
        ka > kb ? 1 /* GT */ : 0 /* EQ */;
}
exports.scalarCompare = scalarCompare;
/**
 * Writes the key and value (by reference) from the given `from` node to the given `to` node.
 *
 * @category Helper
 */
function writeNodeContents(_a) {
    var from = _a.from, to = _a.to;
    to.key = from.key;
    to.value = from.value;
}
exports.writeNodeContents = writeNodeContents;
/**
 * Returns the node within the subtree of given node that ranks lowest.
 * Returns the given node itself if it has no children, or `null`
 * if the given node is itself `null`.
 *
 * @category Tree recursion
 */
function minNode(node) {
    var _a;
    return node === null ? null : (_a = minNode(node.left)) !== null && _a !== void 0 ? _a : node;
}
exports.minNode = minNode;
/**
 * Returns the node within the subtree of given node that ranks highest.
 * Returns the given node itself if it has no children, or `null`
 * if the given node is itself `null`.
 *
 * @category Tree recursion
 */
function maxNode(node) {
    var _a;
    return node === null ? null : (_a = maxNode(node.right)) !== null && _a !== void 0 ? _a : node;
}
exports.maxNode = maxNode;
/**
 * Performs in-order traversal of the given node's subtree, applying the given
 * `fn` to each contained value.
 * @category Tree recursion
 */
function traverseInOrder(node, fn) {
    if (node === null) {
        return;
    }
    traverseInOrder(node.left, fn);
    fn(node);
    traverseInOrder(node.right, fn);
}
exports.traverseInOrder = traverseInOrder;
/**
 * Performs pre-order traversal of the given node's subtree, applying the given
 * `fn` to each contained value.
 * @category Tree recursion
 */
function traversePreOrder(node, fn) {
    if (node === null) {
        return;
    }
    fn(node);
    traversePreOrder(node.left, fn);
    traversePreOrder(node.right, fn);
}
exports.traversePreOrder = traversePreOrder;
/**
 * Performs post-order traversal of the given node's subtree, applying the given
 * `fn` to each contained value.
 * @category Tree recursion
 */
function traversePostOrder(node, fn) {
    if (node === null) {
        return;
    }
    traversePostOrder(node.left, fn);
    traversePostOrder(node.right, fn);
    fn(node);
}
exports.traversePostOrder = traversePostOrder;
/**
 * Folds (reduces) the given node's subtree left-to-right using in-order traversal.
 *
 * @category Tree recursion
 */
function foldLeft(node, fn, seed) {
    if (node === null) {
        return seed;
    }
    seed = foldLeft(node.left, fn, seed);
    seed = fn(seed, node);
    return foldLeft(node.right, fn, seed);
}
exports.foldLeft = foldLeft;
/**
 * Folds (reduces) the given node's subtree right-to-left using reversed in-order traversal.
 *
 * @category Tree recursion
 */
function foldRight(node, fn, seed) {
    if (node === null) {
        return seed;
    }
    seed = foldRight(node.right, fn, seed);
    seed = fn(seed, node);
    return foldRight(node.left, fn, seed);
}
exports.foldRight = foldRight;
/**
 * Search a node by a given `searchKey`, matching against keys of nodes. Returns the
 * matching node if found, or `null` otherwise.
 *
 * @category Tree recursion
 */
function search(node, searchKey) {
    if (node === null) {
        return null;
    }
    switch (scalarCompare(searchKey, node.key)) {
        case -1 /* LT */:
            return search(node.left, searchKey);
        case 0 /* EQ */:
            return node;
        case 1 /* GT */:
            return search(node.right, searchKey);
    }
}
exports.search = search;
/**
 * Inserts the given key an corresponding value into the sub-tree of the given node.
 * Returns an array of length 2 with at:
 *
 * - `[0]` - the node that takes the place of given node after insertion and re-balancing,
 *   which might also be the given node itself.
 *
 * - `[1]` - A boolean indicating whether the new key and value were actually inserted.
 *   This will be `false` if the key was already in the tree prior to insertion.
 *
 * @category Tree recursion
 */
function insert(key, val, node) {
    var _a, _b;
    if (node === null) {
        return [new AVLNode_1.AVLNode(key, val), true];
    }
    var isInserted;
    switch (scalarCompare(key, node.key)) {
        case 0 /* EQ */:
            return [node, false];
        case -1 /* LT */:
            _a = insert(key, val, node.left), node.left = _a[0], isInserted = _a[1];
            break;
        case 1 /* GT */:
            _b = insert(key, val, node.right), node.right = _b[0], isInserted = _b[1];
            break;
    }
    if (!isInserted) {
        return [node, false];
    }
    updateNodeSize(node);
    updateNodeHeight(node);
    return [balanceNode(node), true];
}
exports.insert = insert;
/**
 * Deletes the node by given key from the subtree of given node. Returns an array with:
 *
 * - `[0]` - the node that takes the place of given node after deletion and re-balancing,
 *   which might also be the given node itself.
 *
 * - `[1]` - A boolean indicating whether a node was actually deleted.
 *   This will be `false` if the key was not found in the tree.
 *
 * @category Tree recursion
 */
function deleteKey(key, node) {
    var _a, _b, _c;
    var _d;
    if (node === null) {
        return [null, false];
    }
    var isDeleted;
    switch (scalarCompare(key, node.key)) {
        case -1 /* LT */:
            _a = deleteKey(key, node.left), node.left = _a[0], isDeleted = _a[1];
            break;
        case 1 /* GT */:
            _b = deleteKey(key, node.right), node.right = _b[0], isDeleted = _b[1];
            break;
        case 0 /* EQ */:
            var hasLeft = node.left !== null;
            var hasRight = node.right !== null;
            // Zero child case - can return immediately
            if (neither(hasLeft, hasRight)) {
                return [null, true];
            }
            // One child case - can return immediately
            if (either(hasLeft, hasRight)) {
                return [(_d = node.left) !== null && _d !== void 0 ? _d : node.right, true];
            }
            // Two child case - must re-balance afterwards
            var successor = minNode(node.right);
            writeNodeContents({ from: successor, to: node });
            _c = deleteKey(successor.key, node.right), node.right = _c[0], isDeleted = _c[1];
    }
    updateNodeSize(node);
    updateNodeHeight(node);
    return [balanceNode(node), isDeleted];
}
exports.deleteKey = deleteKey;
/**
 * Balances the given node if it is unbalanced. Returns the node that takes the place of the
 * given node after balancing, which is the given node itself if no balancing is required.
 *
 * @category Helper
 */
function balanceNode(node) {
    var zBalance = nodeBalance(node);
    var lBalance = nodeBalance(node.left);
    var rBalance = nodeBalance(node.right);
    /*────────────────────────────────────────────────────────────────────────────────────────┐
    │                                  Left - Left case                                       │
    ├─────────────────────────────────────────────────────────────────────────────────────────┤
    │                      z                                      y                           │
    │                     / \                                   /   \                         │
    │                    y   T4      Right Rotate (z)          x      z                       │
    │                   / \          - - - - - - - - ->      /  \    /  \                     │
    │                  x   T3                               T1  T2  T3  T4                    │
    │                 / \                                                                     │
    │               T1   T2                                                                   │
    └────────────────────────────────────────────────────────────────────────────────────────*/
    if (zBalance < -1 && lBalance <= 0) {
        return rotateRight(node);
    }
    /*────────────────────────────────────────────────────────────────────────────────────────┐
    │                                 Left - Right case                                       │
    ├─────────────────────────────────────────────────────────────────────────────────────────┤
    │         z                                  z                                x           │
    │        / \                               /   \                             /  \         │
    │       y   T4    Left Rotate (y)         x     T4   Right Rotate(z)       y      z       │
    │      / \        - - - - - - - - ->     /  \        - - - - - - - ->     / \    / \      │
    │    T1   x                             y    T3                         T1  T2  T3  T4    │
    │        / \                           / \                                                │
    │      T2   T3                       T1   T2                                              │
    └────────────────────────────────────────────────────────────────────────────────────────*/
    if (zBalance < -1 && lBalance > 0) {
        node.left = rotateLeft(node.left);
        return rotateRight(node);
    }
    /*────────────────────────────────────────────────────────────────────────────────────────┐
    │                                  Right - Left case                                      │
    ├─────────────────────────────────────────────────────────────────────────────────────────┤
    │      z                                z                                 x               │
    │     / \                              / \                               /  \             │
    │   T1   y     Right Rotate (y)      T1   x        Left Rotate(z)      z      y           │
    │       / \    - - - - - - - - ->       /  \     - - - - - - - ->     / \    / \          │
    │      x   T4                         T2    y                       T1  T2  T3  T4        │
    │     / \                                  /  \                                           │
    │   T2   T3                               T3   T4                                         │
    └────────────────────────────────────────────────────────────────────────────────────────*/
    if (zBalance > 1 && rBalance < 0) {
        node.right = rotateRight(node.right);
        return rotateLeft(node);
    }
    /*────────────────────────────────────────────────────────────────────────────────────────┐
    │                                 Right - Right case                                      │
    ├─────────────────────────────────────────────────────────────────────────────────────────┤
    │                       z                                   y                             │
    │                      /  \                               /   \                           │
    │                    T1    y       Left Rotate(z)        z      x                         │
    │                         /  \     - - - - - - - ->     / \    / \                        │
    │                       T2    x                        T1  T2 T3  T4                      │
    │                             / \                                                         │
    │                           T3   T4                                                       │
    └────────────────────────────────────────────────────────────────────────────────────────*/
    if (zBalance > 1 && rBalance >= 0) {
        return rotateLeft(node);
    }
    return node;
}
/**
 * Performs a left rotation on the given node. Returns the node that takes the
 * place of the given node after rotation.
 *
 * ```text
 * ┌──────────────────────────────────────────────────────────────────┐
 * │ argument -> z                                      y <- return   │
 * │            /  \                                  /   \           │
 * │           T1   y         rotateLeft(z)          z      x         │
 * │               /  \       - - - - - - - ->      / \    / \        │
 * │              T2   x                           T1  T2 T3  T4      │
 * │                  / \                                             │
 * │                T3  T4                                            │
 * └──────────────────────────────────────────────────────────────────┘
 * ```
 * @category Helper
 */
function rotateLeft(node) {
    if (node.right === null) {
        throw new Error('Cannot left-rotate a node without a right child');
    }
    var R = node.right;
    var RL = node.right.left;
    R.left = node;
    node.right = RL;
    updateNodeSize(node);
    updateNodeSize(R);
    updateNodeHeight(node);
    updateNodeHeight(R);
    return R;
}
exports.rotateLeft = rotateLeft;
/**
 * Performs a right rotation on the given node. Returns the node that takes the
 * place of the given node after rotation.
 *
 * ```text
 * ┌──────────────────────────────────────────────────────────────────┐
 * │ argument -> z                                      y <- return   │
 * │            / \                                   /   \           │
 * │           y   T4      rotateRight(z)            x      z         │
 * │          / \          - - - - - - - - ->      /  \    /  \       │
 * │         x   T3                               T1  T2  T3  T4      │
 * │        / \                                                       │
 * │      T1   T2                                                     │
 * └──────────────────────────────────────────────────────────────────┘
 * ```
 * @category Helper
 */
function rotateRight(node) {
    if (node.left === null) {
        throw new Error('Cannot right-rotate a node without a left child');
    }
    var L = node.left;
    var LR = node.left.right;
    L.right = node;
    node.left = LR;
    updateNodeSize(node);
    updateNodeSize(L);
    updateNodeHeight(node);
    updateNodeHeight(L);
    return L;
}
exports.rotateRight = rotateRight;
function either(a, b) {
    return a ? !b : b;
}
function neither(a, b) {
    return !a && !b;
}
