/**
 * This module holds the class implementation of the public API interface
 * {@link IAVLTree `IAVLTree`}.
 *
 * @module Internal - AVLTree
 */
import { deleteKey, foldLeft, foldRight, traverseInOrder, insert, maxNode, minNode, search } from './functions';
/**
 * The class implementation of the public API interface {@link IAVLTree `IAVLTree`}.
 *
 * @typeParam K The key type that is derived from tree values, in order to compare their order.
 * @typeParam V The type of values that are to be stored in the tree.
 */
var AVLTree = /** @class */ (function () {
    function AVLTree(getKey) {
        this.getKey = getKey;
        this.root = null;
    }
    AVLTree.prototype.isEmpty = function () {
        return this.root === null;
    };
    AVLTree.prototype.size = function () {
        var _a, _b;
        return (_b = (_a = this.root) === null || _a === void 0 ? void 0 : _a.size) !== null && _b !== void 0 ? _b : 0;
    };
    AVLTree.prototype.minValue = function () {
        var _a, _b;
        return (_b = (_a = minNode(this.root)) === null || _a === void 0 ? void 0 : _a.value) !== null && _b !== void 0 ? _b : null;
    };
    AVLTree.prototype.maxValue = function () {
        var _a, _b;
        return (_b = (_a = maxNode(this.root)) === null || _a === void 0 ? void 0 : _a.value) !== null && _b !== void 0 ? _b : null;
    };
    AVLTree.prototype.search = function (key) {
        var _a, _b;
        return (_b = (_a = search(this.root, key)) === null || _a === void 0 ? void 0 : _a.value) !== null && _b !== void 0 ? _b : null;
    };
    AVLTree.prototype.forEach = function (fn) {
        traverseInOrder(this.root, function (node) { return fn(node.value); });
    };
    AVLTree.prototype.foldLeft = function (fn, seed) {
        return foldLeft(this.root, function (acc, node) { return fn(acc, node.value); }, seed);
    };
    AVLTree.prototype.foldRight = function (fn, seed) {
        return foldRight(this.root, function (acc, node) { return fn(acc, node.value); }, seed);
    };
    AVLTree.prototype.keys = function () {
        return this.toArray(function (node) { return node.key; });
    };
    AVLTree.prototype.values = function () {
        return this.toArray(function (node) { return node.value; });
    };
    AVLTree.prototype.insert = function (value) {
        var _a = insert(this.getKey(value), value, this.root), rootAfterInsertion = _a[0], isInserted = _a[1];
        this.root = rootAfterInsertion;
        return isInserted;
    };
    AVLTree.prototype.delete = function (key) {
        var _a = deleteKey(key, this.root), rootAfterDeletion = _a[0], isDeleted = _a[1];
        this.root = rootAfterDeletion;
        return isDeleted;
    };
    AVLTree.prototype.toArray = function (transformer) {
        return foldLeft(this.root, function (acc, node) {
            acc.push(transformer(node));
            return acc;
        }, []);
    };
    return AVLTree;
}());
export { AVLTree };
