/*!
 * AVL Binary search trees
 * https://jjwesterkamp.github.io/avl-bst
 * (c) 2021 Jeffrey Westerkamp
 * This software may be freely distributed under the MIT license.
 */
(function webpackUniversalModuleDefinition(root, factory) {
	if(typeof exports === 'object' && typeof module === 'object')
		module.exports = factory();
	else if(typeof define === 'function' && define.amd)
		define([], factory);
	else if(typeof exports === 'object')
		exports["AVLTree"] = factory();
	else
		root["AVLTree"] = factory();
})(self, function() {
return /******/ (() => { // webpackBootstrap
/******/ 	"use strict";
/******/ 	// The require scope
/******/ 	var __webpack_require__ = {};
/******/ 	
/************************************************************************/
/******/ 	/* webpack/runtime/define property getters */
/******/ 	(() => {
/******/ 		// define getter functions for harmony exports
/******/ 		__webpack_require__.d = (exports, definition) => {
/******/ 			for(var key in definition) {
/******/ 				if(__webpack_require__.o(definition, key) && !__webpack_require__.o(exports, key)) {
/******/ 					Object.defineProperty(exports, key, { enumerable: true, get: definition[key] });
/******/ 				}
/******/ 			}
/******/ 		};
/******/ 	})();
/******/ 	
/******/ 	/* webpack/runtime/hasOwnProperty shorthand */
/******/ 	(() => {
/******/ 		__webpack_require__.o = (obj, prop) => (Object.prototype.hasOwnProperty.call(obj, prop))
/******/ 	})();
/******/ 	
/************************************************************************/
var __webpack_exports__ = {};

// EXPORTS
__webpack_require__.d(__webpack_exports__, {
  "default": () => (/* binding */ src)
});

;// CONCATENATED MODULE: ./src/AVLNode.ts
/**
 * This module holds the class implementation of the AVLNode tree nodes.
 *
 * @module Internal - AVLNode
 */
/**
 * This is the class implementation of the AVL tree nodes.
 *
 * @typeParam K The key type of the node.
 * @typeParam V The value type of the node.
 */
var AVLNode = /** @class */ (function () {
    /**
     * @param key The key of the new node.
     * @param value The value of the new node.
     */
    function AVLNode(key, value) {
        this.key = key;
        this.value = value;
        /**
         * The height of the node: 1 plus the height of the greater height child of the node.
         * Nodes without children have a height of 1.
         */
        this.height = 1;
        /**
         * The size of the node: the amount of nodes (including this node itself)
         * within the node's subtree.
         * @type {number}
         */
        this.size = 1;
        /**
         * The left child of the node.
         */
        this.left = null;
        /**
         * The right child of the node.
         */
        this.right = null;
    }
    return AVLNode;
}());


;// CONCATENATED MODULE: ./src/functions.ts
/**
 * The internal recursive functions used by the AVLTree class, and additional
 * helper functions. These perform outside of the tree class context, and take
 * nodes to operate on. This makes them reusable for many more applications.
 *
 * While the AVLTree interface only exposes its contained values the functions
 * here return nodes with these values within, so that internally we can easily
 * traverse subtrees and otherwise operate on them.
 *
 * @module Internal - Functions
 */

/**
 * Returns the height of given node or zero if the node is `null`.
 *
 * @category Helper
 */
function nodeHeight(node) {
    var _a;
    return (_a = node === null || node === void 0 ? void 0 : node.height) !== null && _a !== void 0 ? _a : 0;
}
/**
 * Updates the given node's height property based on the current heights
 * of its left and right subtrees.
 *
 * @category Helper
 */
function updateNodeHeight(node) {
    node.height = 1 + Math.max(nodeHeight(node.left), nodeHeight(node.right));
}
/**
 * Updates the given node's size property based on the current sizes
 * of its left and right subtrees.
 *
 * @category Helper
 */
function updateNodeSize(node) {
    var _a, _b, _c, _d;
    node.size = 1 + ((_b = (_a = node.left) === null || _a === void 0 ? void 0 : _a.size) !== null && _b !== void 0 ? _b : 0) + ((_d = (_c = node.right) === null || _c === void 0 ? void 0 : _c.size) !== null && _d !== void 0 ? _d : 0);
}
/**
 * Returns the given node's balance: the relationship between the heights
 * of its left and right subtrees. Returns a negative number if the node's
 * left child is the higher height child, or a positive number if the right
 * child is the higher height child. Returns zero if either the node is `null`
 * or both its children have equal height.
 *
 * @category Helper
 */
function nodeBalance(node) {
    return node ? nodeHeight(node.right) - nodeHeight(node.left) : 0;
}
/**
 * Takes two values of type `K extends Orderable` and returns their ordering.
 *
 * @category Helper
 */
function scalarCompare(ka, kb) {
    return ka < kb ? -1 /* LT */ :
        ka > kb ? 1 /* GT */ : 0 /* EQ */;
}
/**
 * Writes the key and value (by reference) from the given `from` node to the given `to` node.
 *
 * @category Helper
 */
function writeNodeContents(_a) {
    var from = _a.from, to = _a.to;
    to.key = from.key;
    to.value = from.value;
}
/**
 * Returns the node within the subtree of given node that ranks lowest.
 * Returns the given node itself if it has no children, or `null`
 * if the given node is itself `null`.
 *
 * @category Tree recursion
 */
function minNode(node) {
    var _a;
    return node === null ? null : (_a = minNode(node.left)) !== null && _a !== void 0 ? _a : node;
}
/**
 * Returns the node within the subtree of given node that ranks highest.
 * Returns the given node itself if it has no children, or `null`
 * if the given node is itself `null`.
 *
 * @category Tree recursion
 */
function maxNode(node) {
    var _a;
    return node === null ? null : (_a = maxNode(node.right)) !== null && _a !== void 0 ? _a : node;
}
/**
 * Performs in-order traversal of the given node's subtree, applying the given
 * `fn` to each contained value.
 * @category Tree recursion
 */
function traverseInOrder(node, fn) {
    if (node === null) {
        return;
    }
    traverseInOrder(node.left, fn);
    fn(node);
    traverseInOrder(node.right, fn);
}
/**
 * Performs pre-order traversal of the given node's subtree, applying the given
 * `fn` to each contained value.
 * @category Tree recursion
 */
function traversePreOrder(node, fn) {
    if (node === null) {
        return;
    }
    fn(node);
    traversePreOrder(node.left, fn);
    traversePreOrder(node.right, fn);
}
/**
 * Performs post-order traversal of the given node's subtree, applying the given
 * `fn` to each contained value.
 * @category Tree recursion
 */
function traversePostOrder(node, fn) {
    if (node === null) {
        return;
    }
    traversePostOrder(node.left, fn);
    traversePostOrder(node.right, fn);
    fn(node);
}
/**
 * Folds (reduces) the given node's subtree left-to-right using in-order traversal.
 *
 * @category Tree recursion
 */
function foldLeft(node, fn, seed) {
    if (node === null) {
        return seed;
    }
    seed = foldLeft(node.left, fn, seed);
    seed = fn(seed, node);
    return foldLeft(node.right, fn, seed);
}
/**
 * Folds (reduces) the given node's subtree right-to-left using reversed in-order traversal.
 *
 * @category Tree recursion
 */
function foldRight(node, fn, seed) {
    if (node === null) {
        return seed;
    }
    seed = foldRight(node.right, fn, seed);
    seed = fn(seed, node);
    return foldRight(node.left, fn, seed);
}
/**
 * Search a node by a given `searchKey`, matching against keys of nodes. Returns the
 * matching node if found, or `null` otherwise.
 *
 * @category Tree recursion
 */
function search(node, searchKey) {
    if (node === null) {
        return null;
    }
    switch (scalarCompare(searchKey, node.key)) {
        case -1 /* LT */:
            return search(node.left, searchKey);
        case 0 /* EQ */:
            return node;
        case 1 /* GT */:
            return search(node.right, searchKey);
    }
}
/**
 * Inserts the given key an corresponding value into the sub-tree of the given node.
 * Returns an array of length 2 with at:
 *
 * - `[0]` - the node that takes the place of given node after insertion and re-balancing,
 *   which might also be the given node itself.
 *
 * - `[1]` - A boolean indicating whether the new key and value were actually inserted.
 *   This will be `false` if the key was already in the tree prior to insertion.
 *
 * @category Tree recursion
 */
function insert(key, val, node) {
    var _a, _b;
    if (node === null) {
        return [new AVLNode(key, val), true];
    }
    var isInserted;
    switch (scalarCompare(key, node.key)) {
        case 0 /* EQ */:
            return [node, false];
        case -1 /* LT */:
            _a = insert(key, val, node.left), node.left = _a[0], isInserted = _a[1];
            break;
        case 1 /* GT */:
            _b = insert(key, val, node.right), node.right = _b[0], isInserted = _b[1];
            break;
    }
    if (!isInserted) {
        return [node, false];
    }
    updateNodeSize(node);
    updateNodeHeight(node);
    return [balanceNode(node), true];
}
/**
 * Deletes the node by given key from the subtree of given node. Returns an array with:
 *
 * - `[0]` - the node that takes the place of given node after deletion and re-balancing,
 *   which might also be the given node itself.
 *
 * - `[1]` - A boolean indicating whether a node was actually deleted.
 *   This will be `false` if the key was not found in the tree.
 *
 * @category Tree recursion
 */
function deleteKey(key, node) {
    var _a, _b, _c;
    var _d;
    if (node === null) {
        return [null, false];
    }
    var isDeleted;
    switch (scalarCompare(key, node.key)) {
        case -1 /* LT */:
            _a = deleteKey(key, node.left), node.left = _a[0], isDeleted = _a[1];
            break;
        case 1 /* GT */:
            _b = deleteKey(key, node.right), node.right = _b[0], isDeleted = _b[1];
            break;
        case 0 /* EQ */:
            var hasLeft = node.left !== null;
            var hasRight = node.right !== null;
            // Zero child case - can return immediately
            if (neither(hasLeft, hasRight)) {
                return [null, true];
            }
            // One child case - can return immediately
            if (either(hasLeft, hasRight)) {
                return [(_d = node.left) !== null && _d !== void 0 ? _d : node.right, true];
            }
            // Two child case - must re-balance afterwards
            var successor = minNode(node.right);
            writeNodeContents({ from: successor, to: node });
            _c = deleteKey(successor.key, node.right), node.right = _c[0], isDeleted = _c[1];
    }
    updateNodeSize(node);
    updateNodeHeight(node);
    return [balanceNode(node), isDeleted];
}
/**
 * Balances the given node if it is unbalanced. Returns the node that takes the place of the
 * given node after balancing, which is the given node itself if no balancing is required.
 *
 * @category Helper
 */
function balanceNode(node) {
    var zBalance = nodeBalance(node);
    var lBalance = nodeBalance(node.left);
    var rBalance = nodeBalance(node.right);
    /*────────────────────────────────────────────────────────────────────────────────────────┐
    │                                  Left - Left case                                       │
    ├─────────────────────────────────────────────────────────────────────────────────────────┤
    │                      z                                      y                           │
    │                     / \                                   /   \                         │
    │                    y   T4      Right Rotate (z)          x      z                       │
    │                   / \          - - - - - - - - ->      /  \    /  \                     │
    │                  x   T3                               T1  T2  T3  T4                    │
    │                 / \                                                                     │
    │               T1   T2                                                                   │
    └────────────────────────────────────────────────────────────────────────────────────────*/
    if (zBalance < -1 && lBalance <= 0) {
        return rotateRight(node);
    }
    /*────────────────────────────────────────────────────────────────────────────────────────┐
    │                                 Left - Right case                                       │
    ├─────────────────────────────────────────────────────────────────────────────────────────┤
    │         z                                  z                                x           │
    │        / \                               /   \                             /  \         │
    │       y   T4    Left Rotate (y)         x     T4   Right Rotate(z)       y      z       │
    │      / \        - - - - - - - - ->     /  \        - - - - - - - ->     / \    / \      │
    │    T1   x                             y    T3                         T1  T2  T3  T4    │
    │        / \                           / \                                                │
    │      T2   T3                       T1   T2                                              │
    └────────────────────────────────────────────────────────────────────────────────────────*/
    if (zBalance < -1 && lBalance > 0) {
        node.left = rotateLeft(node.left);
        return rotateRight(node);
    }
    /*────────────────────────────────────────────────────────────────────────────────────────┐
    │                                  Right - Left case                                      │
    ├─────────────────────────────────────────────────────────────────────────────────────────┤
    │      z                                z                                 x               │
    │     / \                              / \                               /  \             │
    │   T1   y     Right Rotate (y)      T1   x        Left Rotate(z)      z      y           │
    │       / \    - - - - - - - - ->       /  \     - - - - - - - ->     / \    / \          │
    │      x   T4                         T2    y                       T1  T2  T3  T4        │
    │     / \                                  /  \                                           │
    │   T2   T3                               T3   T4                                         │
    └────────────────────────────────────────────────────────────────────────────────────────*/
    if (zBalance > 1 && rBalance < 0) {
        node.right = rotateRight(node.right);
        return rotateLeft(node);
    }
    /*────────────────────────────────────────────────────────────────────────────────────────┐
    │                                 Right - Right case                                      │
    ├─────────────────────────────────────────────────────────────────────────────────────────┤
    │                       z                                   y                             │
    │                      /  \                               /   \                           │
    │                    T1    y       Left Rotate(z)        z      x                         │
    │                         /  \     - - - - - - - ->     / \    / \                        │
    │                       T2    x                        T1  T2 T3  T4                      │
    │                             / \                                                         │
    │                           T3   T4                                                       │
    └────────────────────────────────────────────────────────────────────────────────────────*/
    if (zBalance > 1 && rBalance >= 0) {
        return rotateLeft(node);
    }
    return node;
}
/**
 * Performs a left rotation on the given node. Returns the node that takes the
 * place of the given node after rotation.
 *
 * ```text
 * ┌──────────────────────────────────────────────────────────────────┐
 * │ argument -> z                                      y <- return   │
 * │            /  \                                  /   \           │
 * │           T1   y         rotateLeft(z)          z      x         │
 * │               /  \       - - - - - - - ->      / \    / \        │
 * │              T2   x                           T1  T2 T3  T4      │
 * │                  / \                                             │
 * │                T3  T4                                            │
 * └──────────────────────────────────────────────────────────────────┘
 * ```
 * @category Helper
 */
function rotateLeft(node) {
    if (node.right === null) {
        throw new Error('Cannot left-rotate a node without a right child');
    }
    var R = node.right;
    var RL = node.right.left;
    R.left = node;
    node.right = RL;
    updateNodeSize(node);
    updateNodeSize(R);
    updateNodeHeight(node);
    updateNodeHeight(R);
    return R;
}
/**
 * Performs a right rotation on the given node. Returns the node that takes the
 * place of the given node after rotation.
 *
 * ```text
 * ┌──────────────────────────────────────────────────────────────────┐
 * │ argument -> z                                      y <- return   │
 * │            / \                                   /   \           │
 * │           y   T4      rotateRight(z)            x      z         │
 * │          / \          - - - - - - - - ->      /  \    /  \       │
 * │         x   T3                               T1  T2  T3  T4      │
 * │        / \                                                       │
 * │      T1   T2                                                     │
 * └──────────────────────────────────────────────────────────────────┘
 * ```
 * @category Helper
 */
function rotateRight(node) {
    if (node.left === null) {
        throw new Error('Cannot right-rotate a node without a left child');
    }
    var L = node.left;
    var LR = node.left.right;
    L.right = node;
    node.left = LR;
    updateNodeSize(node);
    updateNodeSize(L);
    updateNodeHeight(node);
    updateNodeHeight(L);
    return L;
}
function either(a, b) {
    return a ? !b : b;
}
function neither(a, b) {
    return !a && !b;
}

;// CONCATENATED MODULE: ./src/AVLTree.ts
/**
 * This module holds the class implementation of the public API interface
 * {@link IAVLTree `IAVLTree`}.
 *
 * @module Internal - AVLTree
 */

/**
 * The class implementation of the public API interface {@link IAVLTree `IAVLTree`}.
 *
 * @typeParam K The key type that is derived from tree values, in order to compare their order.
 * @typeParam V The type of values that are to be stored in the tree.
 */
var AVLTree = /** @class */ (function () {
    function AVLTree(getKey) {
        this.getKey = getKey;
        this.root = null;
    }
    AVLTree.prototype.isEmpty = function () {
        return this.root === null;
    };
    AVLTree.prototype.size = function () {
        var _a, _b;
        return (_b = (_a = this.root) === null || _a === void 0 ? void 0 : _a.size) !== null && _b !== void 0 ? _b : 0;
    };
    AVLTree.prototype.minValue = function () {
        var _a, _b;
        return (_b = (_a = minNode(this.root)) === null || _a === void 0 ? void 0 : _a.value) !== null && _b !== void 0 ? _b : null;
    };
    AVLTree.prototype.maxValue = function () {
        var _a, _b;
        return (_b = (_a = maxNode(this.root)) === null || _a === void 0 ? void 0 : _a.value) !== null && _b !== void 0 ? _b : null;
    };
    AVLTree.prototype.search = function (key) {
        var _a, _b;
        return (_b = (_a = search(this.root, key)) === null || _a === void 0 ? void 0 : _a.value) !== null && _b !== void 0 ? _b : null;
    };
    AVLTree.prototype.forEach = function (fn) {
        traverseInOrder(this.root, function (node) { return fn(node.value); });
    };
    AVLTree.prototype.foldLeft = function (fn, seed) {
        return foldLeft(this.root, function (acc, node) { return fn(acc, node.value); }, seed);
    };
    AVLTree.prototype.foldRight = function (fn, seed) {
        return foldRight(this.root, function (acc, node) { return fn(acc, node.value); }, seed);
    };
    AVLTree.prototype.keys = function () {
        return this.toArray(function (node) { return node.key; });
    };
    AVLTree.prototype.values = function () {
        return this.toArray(function (node) { return node.value; });
    };
    AVLTree.prototype.insert = function (value) {
        var _a = insert(this.getKey(value), value, this.root), rootAfterInsertion = _a[0], isInserted = _a[1];
        this.root = rootAfterInsertion;
        return isInserted;
    };
    AVLTree.prototype.delete = function (key) {
        var _a = deleteKey(key, this.root), rootAfterDeletion = _a[0], isDeleted = _a[1];
        this.root = rootAfterDeletion;
        return isDeleted;
    };
    AVLTree.prototype.toArray = function (transformer) {
        return foldLeft(this.root, function (acc, node) {
            acc.push(transformer(node));
            return acc;
        }, []);
    };
    return AVLTree;
}());


;// CONCATENATED MODULE: ./src/index.ts

var factories = {
    create: function (getKey) {
        return new AVLTree(getKey);
    },
    scalar: function () {
        return new AVLTree(function (x) { return x; });
    },
};
/* harmony default export */ const src = (factories);

__webpack_exports__ = __webpack_exports__.default;
/******/ 	return __webpack_exports__;
/******/ })()
;
});
//# sourceMappingURL=avl-bst.js.map